#ifndef VLIT_H
#define VLIT_H

#include <defines.h>

/* \brief Uart speed
 * Character speed of serial line, in bauds
 */
#define VLIT_SPEED	9600

/* \brief Uart read timeout
 * If reader can not get any character form serial line within this time, it returns with error code, in ms. Value can be changed by ctl command VLIT_TIMEOUT. This is default value.
 */
#define VLIT_READ_TIMEOUT	8000


/*! \brief Vlit protocol commands. See VLIT documentation.
 */
#define UMC_CMD_ADR_BROADCAST 0x3E
#define UMC_CMD_ADR_BROADCAST_REPLY 0x7E
#define UMC_CMD_ADR_ASSIGN 0x3C
#define UMC_CMD_ADR_ASSIGN_REPLY 0x7C
#define UMC_CMD_GEN_BROADCAST 0x3F
#define UMC_CMD_GEN_BROADCAST_REPLY 0x7F

// send data TO the WSN
#define UMC_CMD_SEND_DATA 0x06
#define UMC_CMD_SEND_DATA_REPLY 0x46
// receive data FROM the WSN
#define UMC_CMD_RECEIVE_DATA 0x07
#define UMC_CMD_RECEIVE_DATA_REPLY 0x47
// receive data from arbitrary sensor count
#define UMC_CMD_RECEIVE_DATA_ARB 0x08

#define UMC_CMD_NETCMD 0x20
#define UCM_CMD_NETCMD_ANSWER (0x20 | 0x40)

#define VLIT_DATA_COUNT 1
#define DEVICE_NONE 65535
#define UMC_START_BYTE 0x03


/*! \brief Retry count
 * Default number of retries, when accessing VLIT end device. Value can be changed by ctl command VLIT_RETRY.
 */
#define VLIT_RETRY_DEFAULT	5


enum vlit_conf
{
/*! \brief Ctl command - debug enable
 * If set to 1, debugging of line is enabled. Driver will print whole binary communication (input and output) to debug line
 * Note that, this can slow down driver response to VLIT node AP.
 */
	VLIT_DEBUG = 0,
/*! \brief Ctl command - set timeout
 * Set communication timeout in seconds. VLIT sensor network must respond for request within this time otherwise communication is reported as unssucesfull.
 */
	VLIT_TIMEOUT,
/*! \brief Ctl command - retry count
 * Sets number of retries to get desired values before its reported as unreachable.
 */
	VLIT_RETRY
};


/*! \brief Phenomenon types
 * Types of phenomenons VLIT node is measuring. Use as parameter for get function.
 */
enum vlit_types
{
	VLIT_AIR_TEMP = 0,	/*! \brief Air temperature */
	VLIT_AIR_HUMI,		/*! \brief Air humidity */
	VLIT_SOIL_TEMP,		/*! \brief Soil temperature */
	VLIT_SOIL_HUMI,		/*! \brief Soil humidity */
	VLIT_BAT_VOLT,		/*! \brief Battery voltage */
	VLIT_RSSI,			/*! \brief RSSI - link quality */
	VLIT_RETRIES,		/*! \brief Number of retries needed to get last measurement */
};


/*! \brief UMC protocol
 * Defines umc protocol packet. See VLIT documentation for more info.
 */
typedef uint16_t crc_t;

#pragma pack(push,1) // For IAR Compiler

typedef struct umc_header_t
{
    // packet start, always equal to 03h
    uint8_t start;
    // length of packet from CMD byte to last byte before CRC.
    uint8_t len;
    // type of command;
    // in reply command is the same, with sixth bit set to 1 (OR 40h).
    uint8_t cmd;
    // destination address
    uint16_t dst;
    // source address
    uint16_t src;
    // sequence number
    uint8_t sqn;
} PACKED umc_header_t;

typedef struct umc_body_t
{
    uint16_t rf_adr; // address of the sensor device
    uint8_t app_len; // (?) application data length
    uint8_t app_cmd; // (?) application command
} PACKED umc_body_t;


/*! \brief Temporary buffer
 * Structure is used to store previously obtained data. When driver is asked for different phenomenon value and this structure contains valid data. Value from this structure is used instead of asking network again
 */
typedef struct vlit_values_t
{
    uint16_t soil_temp;
    uint16_t soil_humi;
    uint16_t air_temp;
    uint16_t air_humi;
    uint16_t bat_volt;
    uint8_t rssi;
} PACKED vlit_values_t;

typedef struct vlit_data_t
{
    umc_body_t metadata;
    vlit_values_t values;
} PACKED vlit_data_t;

// Packet format for data from arbitrary sensor count
// To be used with UMC_CMD_RECEIVE_DATA_ARB
typedef struct vlit_values_arb_t
{
    uint8_t num; // Number of measurements
    struct {
	uint8_t  type;
	int32_t  val;
    } PACKED values[];
} PACKED vlit_values_arb_t;

#pragma pack(pop)

// ----------------------------------

// MSP430F2252 has only one USART
#define GATEWAY_USART_ID 0
// all of 9600, 19200, 38400, 57600, and 115200 supported
#define GATEWAY_USART_BAUDRATE 9600

void vlitInit(void);

void vlitSend(uint8_t cmd, void *data, uint8_t len, uint16_t sender);

void setNodeData(vlit_values_t *data, uint16_t src);

#endif
