/**
 * Copyright (c) 2011-2013, Institute of Electronics and Computer Science
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of  conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

//#define DEBUG 1
//#define BASESTATION_NODEBUG 1

#include <string.h>

#include <delay.h>
#include <radio.h>
#include <random.h>
#include <alarms.h>
#include <defines.h>
#include <net/address.h>
#include <net/networking.h>
#include <leds.h>
#include <watchdog.h>

#include "simple_mhop.h"

// This is one file for all three configurations (sender, forwarder,
// base station, and not all functions are used everywhere. To prevent
// numerous #ifdefs, we disable the defined-but-unused warnings.
#pragma GCC diagnostic ignored "-Wunused-function"
#pragma GCC diagnostic ignored "-Wunused-variable"


#if defined(DEBUG) && !defined(BASESTATION_NODEBUG)
#  include <lib/dprint.h>
#  define DPRINTF(...) PRINTF(__VA_ARGS__)
#else
#  define DPRINTF(...) ((void)0)
#endif // DEBUG

#ifdef USE_SYSLOG
#  include <lib/dprint.h>
#endif

#ifndef SYSLOG_FILTER
#  define SYSLOG_FILTER LOG_INFO /* Filter INFO and higher */
#endif


#if !defined(USE_ROLE_BASE_STATION) && !defined(USE_ROLE_FORWARDER)
#  define USE_ROLE_SENDER
#endif

#define MAX_HOPS      6
#define SEED_DELAY    (15L * 60 * 1000)

#define ADDR_BCAST    0x0000  // Broadcast address
#define ADDR_BS       MOS_ADDR_BASESTATION

#define MAX_RETRIES   3


enum packetType {
    TYPE_SEED,
    TYPE_DATA,
    TYPE_SYSLOG
};

typedef uint16_t pnum_t;

// We know packet length from radioRecv()
struct packet {
    uint8_t type;
    addr_t  srcAddr, dstAddr, origAddr;
    pnum_t  num;
    uint8_t hops;
    uint8_t data[];
} PACKED;

static recvHandler_t handler;

static union {
    uint8_t d[RADIO_MAX_PACKET];
    struct packet p;
} buf;

#define MAX_PAYLOAD (RADIO_MAX_PACKET - sizeof(struct packet))

static pnum_t counter = 1;

static addr_t  nextHop = ADDR_BS;
static pnum_t  lastSeed;
static uint8_t hopsToBS;

uint16_t periodFromGW;

static inline void longDelay(uint32_t ms)
{
    uint32_t i;
    for (i = 0; i < ms/100; i++)
    {
        mdelay(100);
#ifdef USE_ROLE_BASE_STATION
        watchdogStart(WATCHDOG_EXPIRE_1000MS);
#endif
    }
}

static void randomDelay(void)
{
    switch (randomInRange(0, 5))
    {
#define X(i) case (i): (longDelay((i) * 1300L)); break;

        X(0)
        X(1)
        X(2)
        X(3)
        X(4)
        X(5)
        X(6)

#undef X
    }
}

static inline void printLog(addr_t sender, const uint8_t *data)
{
#ifdef USE_SYSLOG
    static const char *prefix[] = {
        [LOG_CRITICAL]  = "CRIT",
        [LOG_ERROR]     = "ERR ",
        [LOG_WARNING]   = "WARN",
        [LOG_INFO]      = "INFO",
        [LOG_DEBUG]     = "DBG "
    };

    uint8_t pri = data[0];

    if (pri >= SYSLOG_FILTER || pri >= LOG_COUNT)
    {
        return;
    }

    PRINTF("%s[%2u]: %.64s\n", prefix[pri], sender, data+1);
#endif
}

#define R(from, to) if (localAddress == (from)) { buf.p.dstAddr = (to); }
#define DO_STATIC_ROUTING \
    R(0x6A5B, 0x0001) \
    \
    R(0x5DB5, 0x6A5B) \
    R(0x334E, 0x5DB5) \
    R(0x6B02, 0x334E) \
    R(0x0BF2, 0x6B02) \
    \
    R(0x7FA5, 0x6A5B) \
    R(0x6254, 0x7FA5) \
    R(0x69F5, 0x6254) \
    R(0x0CAD, 0x69F5) \
    \
    R(0x5421, 0x6A5B) \
    R(0x734C, 0x5421) \
    R(0x38D3, 0x734C) \
    R(0x0CC6, 0x38D3)


static void forwardPacket(uint8_t len)
{
    uint8_t retries;

    ledOn();

    DPRINTF("Forwarding packet...\n");

    buf.p.srcAddr = localAddress;
    if (buf.p.type != TYPE_SEED)
    {
        buf.p.dstAddr = nextHop;

#if 0
        DO_STATIC_ROUTING
#endif
    }
    buf.p.hops++;

    retries = 0;
    do
    {
        // Delay comes first
        randomDelay();
    }
    while (radioSend(buf.d, len) != 0 && ++retries < MAX_RETRIES);

    ledOff();
}

static void radioRecvFunc(void)
{
    static volatile bool busy;

    int i;

    if (busy)
    {
        radioDiscard();
        return;
    }
    busy = true;

    uint8_t len = radioRecv(buf.d, sizeof(buf.d));

    if (len < sizeof(struct packet) || buf.p.hops > MAX_HOPS ||
        !(buf.p.dstAddr == ADDR_BCAST || buf.p.dstAddr == localAddress))
    {
        DPRINTF("Dropped message");
        if (len >= sizeof(struct packet))
        {
            DPRINTF(" from src=%d", buf.p.srcAddr);
        }
        DPRINTF("\n");
        goto end;
    }

    DPRINTF("Accepted message:");
    DPRINTF(" t=%d", buf.p.type);
    DPRINTF(" src=0x%04X", buf.p.srcAddr);
    DPRINTF(" dst=0x%04X", buf.p.dstAddr);
    DPRINTF(" orig=0x%04X", buf.p.origAddr);
    DPRINTF(" payload=");
    for (i = sizeof(struct packet); i < len; i++)
        DPRINTF(" %02x", buf.d[i] & 0xFF);
    DPRINTF("\n");

#ifndef USE_ROLE_BASE_STATION
    if (buf.p.type == TYPE_SEED)
    {
        DPRINTF("Message type is seed; ");
        if ((lastSeed != buf.p.num || hopsToBS > buf.p.hops)
            // && buf.p.srcAddr+1 == localAddress // For testing
            )
        {
#if defined(USE_TYPE_E) || defined(USE_TYPE_F)
            DPRINTF("seed accepted\n");
            nextHop  = buf.p.srcAddr;
            lastSeed = buf.p.num;
            hopsToBS = buf.p.hops;
#endif

#ifdef USE_TYPE_F
            {
                extern int measurePeriod;
                extern bool forcePeriod;

                const uint16_t *x = (const uint16_t *)buf.p.data;

                if (*x != 0)
                {
                    forcePeriod = true;
                    measurePeriod = *x * 60;
                }
                else if (forcePeriod)
                {
                    forcePeriod = false;
                    measurePeriod = DEFAULT_MEASURE_PERIOD;
                }
            }
#endif

#ifdef USE_ROLE_FORWARDER
            forwardPacket(len);
#endif
        }
        else
        {
            DPRINTF("seed ignored\n");
        }
    }
    else // TYPE_DATA
    {
#ifdef USE_ROLE_FORWARDER
        forwardPacket(len);
#endif
    }
#endif

#ifdef USE_ROLE_BASE_STATION
    if (handler && buf.p.type == TYPE_DATA)
    {
        DPRINTF("Calling user packet handler\n");
        handler(buf.p.data, len - sizeof(struct packet), buf.p.origAddr);
    }
    else if (buf.p.type == TYPE_SYSLOG)
    {
        printLog(buf.p.origAddr, buf.p.data);
    }
#endif

end:
    busy = false;
}

static Alarm_t seedAlarm;

void broadcastSeed(void *unused)
{
    Handle_t h;
    int retries;

    alarmRemove(&seedAlarm);

    //~ DPRINTF("Broadcasting seed...\n");

    ATOMIC_START(h); // So that buf doesn't get overwritten

    buf.p.type    = TYPE_SEED;
    buf.p.srcAddr = buf.p.origAddr = localAddress;
    buf.p.dstAddr = ADDR_BCAST;
    buf.p.num     = counter++;
    buf.p.hops    = 0;
    memcpy(buf.p.data, &periodFromGW, sizeof(periodFromGW));

    retries = 0;
    while (radioSend(buf.d, sizeof(struct packet) + sizeof(periodFromGW)) != 0 &&
           retries < MAX_RETRIES)
    {
        randomDelay();
        retries++;
    }

    ATOMIC_END(h);

    alarmSchedule(&seedAlarm, SEED_DELAY);
}

void netInit(void)
{
    networkingInit(); // Will generate local address

    radioSetReceiveHandle(radioRecvFunc);

    radioOn();

    //~ DPRINTF("Multihop init OK\n");
    //~ DPRINTF("Local address is 0x%04X\n", localAddress);

#ifdef USE_ROLE_BASE_STATION
    alarmInit(&seedAlarm, broadcastSeed, NULL);
    broadcastSeed(NULL);
#endif
}

void netSetRecvHandler(recvHandler_t h)
{
    handler = h;
}

// Call this function with disabled interrupts to prevent accidental
// overwriting of buf
static bool netSendImpl(const void *data, uint8_t len, uint8_t type)
{
    uint8_t  retries;

    if (len > MAX_PAYLOAD)
    {
        return false;
    }

    DPRINTF("Sending user packet...\n");

    buf.p.type    = type;
    buf.p.srcAddr = buf.p.origAddr = localAddress;
    buf.p.dstAddr = nextHop;
    buf.p.num     = counter++;
    buf.p.hops    = 0;

#if 0
    DO_STATIC_ROUTING
#endif

    retries = 0;
    while (radioSendHeader(buf.d, sizeof(struct packet), data, len) != 0 &&
           retries < MAX_RETRIES)
    {
        randomDelay();
        retries++;
    }

    return retries < MAX_RETRIES;
}

bool netSend(const void *data, uint8_t len)
{
    Handle_t h;
    bool     res;

    ATOMIC_START(h);

    res = netSendImpl(data, len, TYPE_DATA);

    ATOMIC_END(h);

    return true;
}

bool syslog(uint8_t pri, const char *msg)
{
    Handle_t h;
    int      r;
    bool     res;

    if (pri >= SYSLOG_FILTER)
    {
        return true;
    }

    ATOMIC_START(h);

    // Not very pretty
    buf.p.data[0] = pri;
    strncpy((char *)(buf.p.data+1), msg, MAX_PAYLOAD-1);
    buf.p.data[MAX_PAYLOAD-1] = '\0';

#ifndef USE_ROLE_BASE_STATION
    res = netSendImpl(buf.p.data,
                      sizeof(pri) + strlen((const char *)(buf.p.data+1)) + 1,
                      TYPE_SYSLOG);
#else
    printLog(ADDR_BS, buf.p.data);
    res = true;
#endif

    ATOMIC_END(h);

    return res;
}
